/* IRCfs - IRC FileServ for *nix.
 * Copyright (C) 2002  Nick 'Zaf' Clifford <zaf@nrc.co.nz>
 * For licensing details, refer to the LICENSE file in the source
 * code directory.
 */
#include "runtime.h"
#include "appconf.h"
#include "str_hash.h"
#include "list.h"
#include "stack.h"
#include "internal.h"

static struct conf_group *handle_group_line(char *linebuf, 
		struct conf_group *cg);
static int handle_equ_line(char *linebuf,struct conf_group *cg);

static char *get_group_line_name(char *linebuf);

int appconf_loadfile(const char *filename)
{
	FILE *fp;
	static char linebuf[512];
	struct conf_group *cg=NULL,*newcg=NULL;
	struct stack *st;
	int linenum = 0;
	int rv,len;
	struct conf_value *cv;
	char *cp;

	appconf_init();

	fp = fopen(filename,"r");
	if (!fp)
		return -1;

	cg = conf_root_group;
	if (cg == NULL)
		cg = conf_create_group(NULL,"");

	st = stack_create(NULL);
	stack_push(st,cg);
	rv = 0;
	while(fgets(linebuf,sizeof(linebuf)-1,fp) != NULL) {
		linenum++;
		linebuf[sizeof(linebuf)-1] = 0;
		if (*linebuf == '#')
			continue;
		len = strlen(linebuf);
		/* Trim trailing \n and \r */
		while (linebuf[len-1] == '\n' || linebuf[len-1] == '\r')
			linebuf[--len] = 0;
		
		trim_comments(linebuf);
		if (*linebuf == 0)
			continue;
		debug(conf_debug,"Parsing line %s",linebuf);
		
		if (*linebuf == '<' &&
				(*(linebuf+1) != '/')) {
			/* Open of tag */
			newcg = handle_group_line(linebuf,cg);
			if (newcg == NULL) {
				rv = -1;
				break;
			}
			cv = conf_new_group_val(conf_pool,newcg);
			str_hash_set(cg->settings, newcg->name, cv,NULL);
			cg = newcg;
			stack_push(st,cg);
			continue;
		} else if (*linebuf == '<') {
			/* Close of tag */
			cp = get_group_line_name(linebuf);
			if (newcg == NULL) {
				rv = -1;
				break;
			}
			do {
				cg = stack_pop(st);
				debug(conf_debug,"Popped %s",cg ? cg->name :
						"<null>");
			} while (cg && strcasecmp(cg->name,cp));
			cg = stack_peek(st);
			if (!cg) {
				error(conf_debug,"Close of non existant tag "
						"(%s) "
						"in config file %s:%d",
						newcg->name,
						filename, linenum);
				rv = -1;
				break;
			}
			debug(conf_debug,"Now in group %s",
					cg->name);
			continue;
		}
		if (handle_equ_line(linebuf,cg) == -1) {
			rv = -1;
			break;
		}
	}

	conf_print_settings();

	fclose(fp);
	stack_destroy(st);
	return rv;
}
static char *get_group_line_name(char *linebuf)
{
	char *cp,*start;
	cp = linebuf;
	if (*cp == '<')
		cp++;
	if (*cp == '/') {
		cp++;
	}
	start = cp;
	while(*cp && *cp != '>')
		cp++;
	*cp = 0;
	return start;
}
/* Parsers the <group> line, takes current group, returns new group. */
static struct conf_group *handle_group_line(char *linebuf, 
		struct conf_group *cg)
{
	struct conf_group *newcg;
	char *cp;
	cp = get_group_line_name(linebuf);

	newcg = conf_open_group(cg,cp);
	if (newcg == NULL) {
		debug(conf_debug,"Can't find group named %s",cp);
		return NULL;
	}
	return newcg;
}
static int handle_equ_line(char *linebuf,struct conf_group *cg)
{
	char *value,*name,quote=0,*cp;
	struct conf_value *cv;

	ASSERT(cg);
	
	value = name = linebuf;
	while(*value && *value != '=')
		value++;
	if (*value) {
		*value = 0;
		trim(name);
		value++;
		while(isspace(*value))
			value++;
	}
	if (*value == '\'' || *value == '"') {
		quote = *value;
		value++;
		for(cp = value; *cp != quote && *cp != 0; cp++) {
			if (*cp == '\\' && *(cp+1) != 0)
				cp++;
		}
		if (*cp != quote) {
			warn(conf_debug,"Setting %s=%s is missing end quote",
					name,value-1);
		} else {
			*cp = 0;
		}
	}
	debug(conf_debug,"Found %s=%s",name,value);
	
	cv = conf_new_str(conf_pool,value);
	str_hash_set(cg->settings, name,cv,NULL);
	return 0;
}


